@@ -3,6 +3,7 @@ import enum
 import json
 import math
 import shutil
+import socket
 import time
 import unittest
 import unittest.mock
@@ -12,11 +13,12 @@ import http.client
 from pathlib import Path
 
 import aiohttp
-from selenium import webdriver  # type: ignore
-import selenium.webdriver.firefox.options  # type: ignore
-from selenium.webdriver.common.alert import Alert  # type: ignore
-from selenium.webdriver.common.keys import Keys  # type: ignore
-from selenium.webdriver import ActionChains  # type: ignore
+from selenium import webdriver
+import selenium.webdriver.firefox.options
+from selenium.webdriver.common.alert import Alert
+from selenium.webdriver.common.by import By
+from selenium.webdriver.common.keys import Keys
+from selenium.webdriver import ActionChains
 
 import shc.web
 import shc.web.widgets
@@ -64,6 +66,7 @@ class AbstractWebTest(unittest.TestCase):
     @classmethod
     def tearDownClass(cls) -> None:
         cls.driver.close()
+        cls.driver.quit()
 
 
 class SimpleWebTest(AbstractWebTest):
@@ -87,10 +90,10 @@ class SimpleWebTest(AbstractWebTest):
         self.assertIn('Home Page', self.driver.page_source)
         self.assertIn('Home Page', self.driver.title)
         self.assertIn('Another segment', self.driver.page_source)
-        button = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "Foobar"]')
-        self.assertIn("My button group", button.find_element_by_xpath('../..').text)
-        button = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "Bar"]')
-        self.assertIn("Another button group", button.find_element_by_xpath('../..').text)
+        button = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "Foobar"]')
+        self.assertIn("My button group", button.find_element(By.XPATH, '../..').text)
+        button = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "Bar"]')
+        self.assertIn("Another button group", button.find_element(By.XPATH, '../..').text)
 
     def test_main_menu(self) -> None:
         self.server.page('index', menu_entry="Home", menu_icon='home')
@@ -100,21 +103,21 @@ class SimpleWebTest(AbstractWebTest):
         self.driver.get("http://localhost:42080")
 
         # Only search for the (visible) main navigation bar, instead of the hidden sidebare for mobile screens:
-        container = self.driver.find_element_by_css_selector('.pusher')
+        container = self.driver.find_element(By.CSS_SELECTOR, '.pusher')
 
-        home_link = container.find_element_by_css_selector('i.home.icon').find_element_by_xpath('..')
+        home_link = container.find_element(By.CSS_SELECTOR, 'i.home.icon').find_element(By.XPATH, '..')
         self.assertIn("Home", home_link.text)
         home_link.click()
 
-        container = self.driver.find_element_by_css_selector('.pusher')
-        submenu = container.find_element_by_xpath('.//div[contains(text(), "Foo")]')
-        submenu_entry = submenu.find_element_by_xpath('.//a[contains(@class, "item")]')
+        container = self.driver.find_element(By.CSS_SELECTOR, '.pusher')
+        submenu = container.find_element(By.XPATH, './/div[contains(text(), "Foo")]')
+        submenu_entry = submenu.find_element(By.XPATH, './/a[contains(@class, "item")]')
         self.assertFalse(submenu_entry.is_displayed())
         submenu.click()
         self.assertIn("Bar", submenu_entry.text)
         self.assertTrue(submenu_entry.is_displayed())
         self.assertEqual(submenu_entry.get_attribute('href').strip(), "http://localhost:42080/page/another_page/")
-        submenu_entry.find_element_by_css_selector('i.bars.icon')
+        submenu_entry.find_element(By.CSS_SELECTOR, 'i.bars.icon')
 
 
 class MonitoringTest(unittest.TestCase):
@@ -181,16 +184,16 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            checkbox_element = self.driver.find_element_by_xpath(
-                '//*[normalize-space(text()) = "Main Power"]/..//input')
+            checkbox_element = self.driver.find_element(
+                By.XPATH, '//*[normalize-space(text()) = "Main Power"]/..//input')
             self.assertTrue(checkbox_element.is_selected())
 
-            asyncio.run_coroutine_threadsafe(switch_widget.write(False, [self]), loop=self.server_runner.loop).result()
+            self.server_runner.run_coro(switch_widget.write(False, [self]))
             time.sleep(0.05)
             publish_mock.reset_mock()
             self.assertFalse(checkbox_element.is_selected())
 
-            checkbox_container = checkbox_element.find_element_by_xpath('./..')
+            checkbox_container = checkbox_element.find_element(By.XPATH, './..')
             checkbox_container.click()
             time.sleep(0.05)
             self.assertTrue(checkbox_element.is_selected())
@@ -207,8 +210,8 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            checkbox_container = self.driver.find_element_by_xpath(
-                '//*[normalize-space(text()) = "Some Switch"]/..//input/..')
+            checkbox_container = self.driver.find_element(
+                By.XPATH, '//*[normalize-space(text()) = "Some Switch"]/..//input/..')
 
             # Setting to true requires confirmation
             checkbox_container.click()
@@ -245,10 +248,10 @@ class WebWidgetsTest(AbstractWebTest):
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
 
-            b1_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B1"]')
-            b2_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B2"]')
-            b3_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B3"]')
-            b4_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B4"]')
+            b1_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B1"]')
+            b2_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B2"]')
+            b3_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B3"]')
+            b4_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B4"]')
 
             # Check initial states
             self.assertIn('yellow', b1_element.get_attribute('class'))
@@ -256,14 +259,14 @@ class WebWidgetsTest(AbstractWebTest):
             self.assertIn('red', b4_element.get_attribute('class'))
 
             # Check state updates
-            asyncio.run_coroutine_threadsafe(b1.write(False, [self]), loop=self.server_runner.loop).result()
+            self.server_runner.run_coro(b1.write(False, [self]))
             time.sleep(0.05)
             self.assertNotIn('yellow', b1_element.get_attribute('class'))
             self.assertIn('blue', b2_element.get_attribute('class'))
             self.assertIn('red', b4_element.get_attribute('class'))
 
-            asyncio.run_coroutine_threadsafe(b2.write(False, [self]), loop=self.server_runner.loop).result()
-            asyncio.run_coroutine_threadsafe(b4.write(56, [self]), loop=self.server_runner.loop).result()
+            self.server_runner.run_coro(b2.write(False, [self]))
+            self.server_runner.run_coro(b4.write(56, [self]))
             time.sleep(0.05)
             b1_publish.reset_mock()
             b3_publish.reset_mock()
@@ -302,7 +305,7 @@ class WebWidgetsTest(AbstractWebTest):
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
 
-            button_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B1"]')
+            button_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B1"]')
 
             # Check click with alert
             self.assertNotIn('yellow', button_element.get_attribute('class'))
@@ -342,10 +345,11 @@ class WebWidgetsTest(AbstractWebTest):
         self.server_runner.start()
         self.driver.get("http://localhost:42080")
         time.sleep(0.4)
-        value_element = self.driver.find_element_by_xpath('//*[normalize-space(text()) = "Brightness"]/..//*[@data-id]')
+        value_element = self.driver.find_element(
+            By.XPATH, '//*[normalize-space(text()) = "Brightness"]/..//*[@data-id]')
         self.assertEqual("42 lux", value_element.text.strip())
 
-        asyncio.run_coroutine_threadsafe(text_widget.write(56, [self]), loop=self.server_runner.loop).result()
+        self.server_runner.run_coro(text_widget.write(56, [self]))
         time.sleep(0.05)
         self.assertEqual("56 lux", value_element.text.strip())
 
@@ -358,10 +362,10 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            input_element = self.driver.find_element_by_xpath('//*[normalize-space(text()) = "Brightness"]/..//input')
+            input_element = self.driver.find_element(By.XPATH, '//*[normalize-space(text()) = "Brightness"]/..//input')
             self.assertEqual("42", input_element.get_attribute("value"))
 
-            asyncio.run_coroutine_threadsafe(input_widget.write(56, [self]), loop=self.server_runner.loop).result()
+            self.server_runner.run_coro(input_widget.write(56, [self]))
             time.sleep(0.05)
             publish_mock.reset_mock()
             self.assertEqual("56", input_element.get_attribute("value"))
@@ -394,12 +398,11 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            input_element = self.driver.find_element_by_xpath(
-                '//*[normalize-space(text()) = "Message of the Day"]/..//input')
+            input_element = self.driver.find_element(
+                By.XPATH, '//*[normalize-space(text()) = "Message of the Day"]/..//input')
             self.assertEqual("Hello, World!", input_element.get_attribute("value"))
 
-            asyncio.run_coroutine_threadsafe(input_widget.write("Foobar", [self]), loop=self.server_runner.loop)\
-                .result()
+            self.server_runner.run_coro(input_widget.write("Foobar", [self]))
             time.sleep(0.05)
             publish_mock.reset_mock()
             self.assertEqual("Foobar", input_element.get_attribute("value"))
@@ -419,10 +422,10 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            container_element = self.driver.find_element_by_xpath(
-                '//*[normalize-space(text()) = "Amount of Foo"]/../..')
-            slider_element = container_element.find_element_by_css_selector('.slider')
-            handle_element = slider_element.find_element_by_css_selector(".thumb")
+            container_element = self.driver.find_element(
+                By.XPATH, '//*[normalize-space(text()) = "Amount of Foo"]/../..')
+            slider_element = container_element.find_element(By.CSS_SELECTOR, '.slider')
+            handle_element = slider_element.find_element(By.CSS_SELECTOR, ".thumb")
 
             # Center of handle should be somewhere near 30% of width of
             slider_width = slider_element.rect['width']
@@ -456,12 +459,12 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            foo_row_element = self.driver.find_element_by_xpath('//*[normalize-space(text()) = "Foo"]')
-            bar_row_element = self.driver.find_element_by_xpath('//*[normalize-space(text()) = "Bar"]')
+            foo_row_element = self.driver.find_element(By.XPATH, '//*[normalize-space(text()) = "Foo"]')
+            bar_row_element = self.driver.find_element(By.XPATH, '//*[normalize-space(text()) = "Bar"]')
             # Foobar row should not be findable, since Selenium's xpath can only find DOM elements by text content if
             # they are visible
-            foobar_row_element = self.driver.find_element_by_xpath('//*[contains(text(), "Foobar")]')
-            button = foo_row_element.find_element_by_css_selector("button")
+            foobar_row_element = self.driver.find_element(By.XPATH, '//*[contains(text(), "Foobar")]')
+            button = foo_row_element.find_element(By.CSS_SELECTOR, "button")
 
             self.assertTrue(foo_row_element.is_displayed())
             self.assertTrue(bar_row_element.is_displayed())
@@ -469,7 +472,7 @@ class WebWidgetsTest(AbstractWebTest):
             self.assertIn('red', foo_row_element.get_attribute('class'))
             self.assertIn('blue', bar_row_element.get_attribute('class'))
             self.assertIn('yellow', foobar_row_element.get_attribute('class'))
-            self.assertIn('power off', button.find_element_by_css_selector('.icon').get_attribute('class'))
+            self.assertIn('power off', button.find_element(By.CSS_SELECTOR, '.icon').get_attribute('class'))
 
             # Click the button
             button.click()
@@ -477,9 +480,9 @@ class WebWidgetsTest(AbstractWebTest):
             publish_mock.assert_called_once_with(True, unittest.mock.ANY)
 
             # Swap the rows
-            asyncio.run_coroutine_threadsafe(foo_row.write(False, [self]), loop=self.server_runner.loop).result()
-            asyncio.run_coroutine_threadsafe(bar_row.write(False, [self]), loop=self.server_runner.loop).result()
-            asyncio.run_coroutine_threadsafe(foobar_row.write(True, [self]), loop=self.server_runner.loop).result()
+            self.server_runner.run_coro(foo_row.write(False, [self]))
+            self.server_runner.run_coro(bar_row.write(False, [self]))
+            self.server_runner.run_coro(foobar_row.write(True, [self]))
             time.sleep(1)
             self.assertFalse(foo_row_element.is_displayed())
             self.assertFalse(bar_row_element.is_displayed())
@@ -495,10 +498,10 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            wheel_element = self.driver.find_element_by_css_selector('.IroWheel')
-            wheel_handle_element = wheel_element.find_element_by_css_selector('.IroHandle>circle')
-            slider_element = self.driver.find_element_by_css_selector('.IroSlider')
-            slider_handle_element = slider_element.find_element_by_css_selector('.IroHandle>circle')
+            wheel_element = self.driver.find_element(By.CSS_SELECTOR, '.IroWheel')
+            wheel_handle_element = wheel_element.find_element(By.CSS_SELECTOR, '.IroHandle>circle')
+            slider_element = self.driver.find_element(By.CSS_SELECTOR, '.IroSlider')
+            slider_handle_element = slider_element.find_element(By.CSS_SELECTOR, '.IroHandle>circle')
 
             # The wheel handle should be in the center of the wheel ...
             wheel_rect = wheel_element.rect
@@ -552,18 +555,18 @@ class WebWidgetsTest(AbstractWebTest):
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
-            container_element = self.driver.find_element_by_xpath(
-                '//*[normalize-space(text()) = "Select the Foo"]/..')
-            menu_element = container_element.find_element_by_css_selector('.selection.dropdown')
+            container_element = self.driver.find_element(
+                By.XPATH, '//*[normalize-space(text()) = "Select the Foo"]/..')
+            menu_element = container_element.find_element(By.CSS_SELECTOR, '.selection.dropdown')
 
             menu_element.click()
             second_option_element = menu_element\
-                .find_element_by_xpath('.//*[normalize-space(text()) = "SOME_OTHER_VALUE"][contains(@class, "item")]')
+                .find_element(By.XPATH, './/*[normalize-space(text()) = "SOME_OTHER_VALUE"][contains(@class, "item")]')
             self.assertIn("selected", second_option_element.get_attribute('class'))
 
             # select the third option
             third_option_element = menu_element\
-                .find_element_by_xpath('.//*[normalize-space(text()) = "YET_ANOTHER_VALUE"]')
+                .find_element(By.XPATH, './/*[normalize-space(text()) = "YET_ANOTHER_VALUE"]')
             third_option_element.click()
 
             time.sleep(0.05)
@@ -587,7 +590,7 @@ class WebWidgetsTest(AbstractWebTest):
         time.sleep(0.6)
 
         # Check that the background image is served and loaded correctly
-        background_image = self.driver.find_element_by_css_selector('.shc.image-container .background')
+        background_image = self.driver.find_element(By.CSS_SELECTOR, '.shc.image-container .background')
         self.assertTrue(self.driver.execute_script(
             "return arguments[0].complete "
             "&& typeof arguments[0].naturalWidth != \"undefined\" "
@@ -595,7 +598,7 @@ class WebWidgetsTest(AbstractWebTest):
         background_image_rect = background_image.rect
 
         # Check that there is a correctly styled button element, labeled "B1", at the right position
-        b1_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B1"]')
+        b1_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B1"]')
         self.assertIn('yellow', b1_element.get_attribute('class'))
         b1_rect = b1_element.rect
         self.assertAlmostEqual(background_image_rect['x'] + background_image_rect['width'] * 0.3,
@@ -604,7 +607,7 @@ class WebWidgetsTest(AbstractWebTest):
                                b1_rect['y'] + b1_rect['height']/2, delta=4)  # 4px off is okay
 
         # Check that there is a label in the right position
-        l2_element = self.driver.find_element_by_css_selector('.shc.image-container .ui.label')
+        l2_element = self.driver.find_element(By.CSS_SELECTOR, '.shc.image-container .ui.label')
         self.assertEqual("15.3", l2_element.text.strip())
         self.assertIn('red', l2_element.get_attribute('class'))
         l2_rect = l2_element.rect
@@ -655,6 +658,8 @@ class TestAPI(unittest.TestCase):
             await api_object.write(value, [self])
 
         tic = time.time()
+        # The `scheduled_update()` coroutine shall run in parallel with our request below. Thus, we cannot use
+        # InterfaceThreadRunner.run_coro(), which waits for the coroutine to return.
         asyncio.run_coroutine_threadsafe(scheduled_update(56), self.server_runner.loop)
 
         response: http.client.HTTPResponse = urllib.request.urlopen(
@@ -686,6 +691,8 @@ class TestAPI(unittest.TestCase):
 
         # A GET request with matching ETag and wait parameter. It should wait for the next value
         tic = time.time()
+        # The `scheduled_update()` coroutine shall run in parallel with our request below. Thus, we cannot use
+        # InterfaceThreadRunner.run_coro(), which waits for the coroutine to return.
         asyncio.run_coroutine_threadsafe(scheduled_update(56), self.server_runner.loop)
 
         request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object?wait=0.5",
@@ -708,6 +715,33 @@ class TestAPI(unittest.TestCase):
         self.assertEqual(etag2, response.headers['ETag'])
         self.assertAlmostEqual(0, toc-tic, delta=0.02)
 
+    def test_rest_get_wait_abort(self) -> None:
+        """ Test, that an aborted long poll to the API does not destroy any internal state """
+        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
+        self.server_runner.start()
+
+        async def scheduled_update(value):
+            await asyncio.sleep(0.2)
+            await api_object.write(value, [self])
+
+        request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object?wait=0.5")
+        with self.assertRaises(socket.timeout):
+            urllib.request.urlopen(request, timeout=0.25)
+
+        # Check that everything still works:
+        self.server_runner.run_coro(scheduled_update(56))
+
+        tic = time.time()
+        # The `scheduled_update()` coroutine shall run in parallel with our request below. Thus, we cannot use
+        # InterfaceThreadRunner.run_coro(), which waits for the coroutine to return.
+        asyncio.run_coroutine_threadsafe(scheduled_update(128), self.server_runner.loop)
+
+        response: http.client.HTTPResponse = urllib.request.urlopen(
+            "http://localhost:42080/api/v1/object/the_api_object?wait=0.5")
+        toc = time.time()
+        self.assertEqual(128, json.loads(response.read()))
+        self.assertAlmostEqual(0.2, toc-tic, delta=0.02)
+
     def test_rest_post(self) -> None:
         api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
         self.server_runner.start()
@@ -893,8 +927,8 @@ class WebSocketAPITest(unittest.TestCase):
         self.assertNotIn('value', data)
 
         self.ws_callback.reset_mock()
-        asyncio.run_coroutine_threadsafe(api_object.write(56, [self]), self.server_runner.loop)
-        asyncio.run_coroutine_threadsafe(api_object2.write(56, [self]), self.server_runner.loop)
+        self.server_runner.run_coro(api_object.write(56, [self]))
+        self.server_runner.run_coro(api_object2.write(56, [self]))
         await asyncio.sleep(0.05)
 
         self.assertEqual(2, self.ws_callback.call_count)
@@ -902,3 +936,42 @@ class WebSocketAPITest(unittest.TestCase):
         self.assertIn(data['name'], ('the_api_object', 'the_other_api_object'))
         self.assertEqual(56, data['value'])
         self.assertEqual(200, data['status'])
+
+    @async_test
+    async def test_last_will(self) -> None:
+        api_object = self.server.api(int, "the_api_object")
+        self.server_runner.start()
+        await self.start_websocket()
+
+        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
+            # First lastwill request is broken (wrong value type)
+            await self.ws.send_json({'action': 'lastwill', 'name': 'the_api_object', 'value': "hello, world!"})
+            await asyncio.sleep(0.05)
+
+            publish_mock.assert_not_called()
+
+            self.ws_callback.assert_called_once()
+            data = json.loads(self.ws_callback.call_args[0][0])
+            self.assertEqual('lastwill', data['action'])
+            self.assertEqual('the_api_object', data['name'])
+            self.assertEqual(422, data['status'])
+            self.ws_callback.reset_mock()
+
+            # Now, a correct lastwill request. It should still not publish any value at the server
+            await self.ws.send_json({'action': 'lastwill', 'name': 'the_api_object', 'value': 56})
+            await asyncio.sleep(0.05)
+
+            publish_mock.assert_not_called()
+
+            self.ws_callback.assert_called_once()
+            data = json.loads(self.ws_callback.call_args[0][0])
+            self.assertEqual('lastwill', data['action'])
+            self.assertEqual('the_api_object', data['name'])
+            self.assertEqual(204, data['status'])
+
+            # Now, let the client disconnect. This should trigger publishing the value at the server
+            await self.ws.close()
+            await self.client_session.close()
+            await asyncio.sleep(0.1)
+
+            publish_mock.assert_called_once_with(56, unittest.mock.ANY)
